/* FGHC -> FLENG compiler (baseline version)

  Copyright (c) 2021, Felix L. Winkelmann
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. The name of the authors may not be used to endorse or promote products
     derived from this software without specific prior written permission.
  
  THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

:- dynamic([declaration/1, library_mode/0, no_warnings/0, include_path/1, uses/1]).
:- dynamic([moded/2]).
:- initialization(main).

:- if(current_prolog_flag(dialect, gprolog)).
% GNU
:- include('version.pl').
main :-
    argument_list(Args),
    main(Args).
:- endif.

:- if(current_prolog_flag(dialect, swi)).
% SWI
:- include(fleng('version.pl')).

format_to_atom(Out, Fstr, Args) :-
    format_to_codes(Fstr, Args, S),
    atom_codes(Out, S).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.

:- if(current_prolog_flag(dialect, yap)).
% YAP
:- use_module(library(lists)).
:- use_module(library(charsio)).
:- include('version.pl').

format_to_atom(Out, Fstr, Args) :-
    format_to_chars(Fstr, Args, S),
    atom_codes(Out, S).

save_to_file :- save('fghc2fl.yap'), halt(0).

start :- 
    set_prolog_flag(syntax_errors, error),
    set_prolog_flag(unknown, error),
    unix(argv(Args)), 
    catch(main(Args), E, quit(E)).

quit(E) :- 
    print_message(error, E),
    halt(1).

numbervars(Term) :- numbervars(Term, 0, _).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.

main(['-v']) :-
    fleng_version(V),
    format('fghc2fl ~a\n', [V]),
    halt(0).
main(['-I', Dir|Args]) :-
    asserta(include_path(Dir)),
    main(Args).
main(['-w'|Args]) :-
    assertz(no_warnings),
    main(Args).
main([Infile, Outfile]) :-
    op(200, fx, !),
    op(700, xfx, '\\=:='),
    op(500, yfx, '><'),
    op(400, yfx, '\\\\'),
    op(700, yfx, '@'),
    op(900, xfy, '&'),
    op(990, xfx, ':='),
    op(0, xfx, '^'),
    op(0, fx, initialization),
    set_prolog_flag(double_quotes, codes),
    assertz(include_path('.')),
    g_assign(counter, 1),
    see(Infile),
    read_program(Prg),
    seen,
    collect_groups(Prg, Groups),
    translate(Groups, TG),
    open(Outfile, write, Out),
    fleng_version(V),
    format(Out, '%% GENERATED BY fghc2fl [~a] FROM ~a\n\n', [V, Infile]),
    write_declarations(Out),
    write_translation(TG, Out),
    format(Out, '%% END OF FILE\n', []),
    close(Out),
    halt(0).
main(_) :- 
    format(user_error, 'usage: fghc2fl [-v] [-w] INFILE OUTFILE\n', []),
    halt(1).

dbg(X) :- format('~q\n', [X]).
dbg(F, X) :- format(F, [X]).

error(Fstr, Args) :-
    format(user_error, Fstr, Args),
    halt(1).

warning(Fstr, Args) :-
    \+no_warnings,
    format(user_error, 'Warning: ', []),
    format(user_error, Fstr, Args).
warning(_, _).

read_program(PRG) :-
    read_term(TERM, [singletons(SL)]),
    report_singletons(TERM, SL, TERM),
    numbervars(TERM),
    read_program(TERM, PRG).

read_program(end_of_file, []).
read_program(-DECL, PRG) :-
    process_declaration(DECL, PRG).
read_program(TERM, [C|PRG]) :-
    canonicalize(TERM, NA, Args, Guards, Body),
    apply_modes(NA, Args, Guards, Body, C),
    read_program(PRG).

apply_modes(NA, Args1, Guards, Body1, c(NA, Args, Guards, Body)) :-
    moded(NA, Modes),
    !,
    extract_variables((Args1, Guards, Body1), Vars),
    apply_modes2(Args1, Modes, Vars, Args, Body1, Body).
apply_modes(NA, Args, Guards, Body, c(NA, Args, Guards, Body)).    

apply_modes2([], _, _, [], B, B).
apply_modes2([X|Args], ['^'|Modes], Vars, ['$VAR'(V)|Args2], B1, B) :-
    available_variable(Vars, V),
    apply_modes2(Args, Modes, [V|Vars], Args2, (B1, '$VAR'(V) = X), B).
apply_modes2([X|Args], ['?'|Modes], Vars, [X|Args2], B1, B) :-
    apply_modes2(Args, Modes, Vars, Args2, B1, B).

report_singletons(_, [], _).
report_singletons(-_, _, _).
report_singletons(_, SL, _) :-
    format(user_error, 'Warning: singleton variables ', []),
    member(Name = _, SL),
    format(user_error, '~a ', [Name]),
    fail.
report_singletons(_, _, Term) :-
    format(user_error, 'in term:\n\n  ~q\n', [Term]).

process_declaration(library_mode, PRG) :-
    assertz(library_mode),
    read_program(PRG).
process_declaration(include(Fname), PRG) :-
    atom(Fname),
    current_input(In),
    find_include_file(Fname, Fname2),
    see(Fname2),
    read_program(Included),
    seen,
    set_input(In),
    append(Included, Next, PRG),
    read_program(Next).
process_declaration(uses(Mods), PRG) :-
    process_uses(Mods),
    read_program(PRG).
process_declaration(mode(Mode), PRG) :-
    process_mode(Mode),
    read_program(PRG).
process_declaration(provide(Spec, Provider), PRG) :-
    synthesize_provider(Spec, Provider, Code),
    read_program(Code, PRG).
process_declaration(machine(M), PRG) :- 
    check_machine_declaration(M),
    read_program(PRG).
process_declaration(Decl, PRG) :-
    assertz(declaration(Decl)),
    read_program(PRG).

check_machine_declaration(ring).
check_machine_declaration(torus).
check_machine_declaration(M) :-
    error('unsupported machine topology: ~w\n', [M]).

process_mode(T) :-
    T =.. [Name|Modes],
    length(Modes, A),
    assertz(moded(Name/A, Modes)).
process_mode(T) :-
    error('invalid mode declaration: ~w\n', [T]).

process_uses([]) :- !.
process_uses(Mod) :-
    atom(Mod),
    process_uses([Mod]).
process_uses([Mod|Mods]) :-
    atom(Mod),
    add_use(Mod),
    process_uses(Mods).
process_uses([Mod|_]) :-
    error('invalid module specification in `uses\' declaration: ~w\n', [Mod]).

add_use(X) :- \+atom(X), !.
add_use(Mod) :- uses(Mod), !.
add_use(Mod) :- assertz(uses(Mod)).

maybe_add_use(Mod:_) :- !, add_use(Mod).
maybe_add_use(_).

synthesize_provider(Spec, Provider, Code) :-
    Spec =.. [Name|Args],
    parse_provider_args(Args, Derefs, 0, Vars, Params),
    parse_provider(Provider, Params, Caller),
    Head =.. [Name|Params],
    available_variable(Vars, V),
    Code = (Head :- deref(Derefs, '$VAR'(V)), when('$VAR'(V), Caller)).

parse_provider_args([], [], _, [], []). 
parse_provider_args(['?'|Args], ['$VAR'(I)|Derefs], I, [I|Vars], ['$VAR'(I)|Params]) :-
    I2 is I + 1,
    parse_provider_args(Args, Derefs, I2, Vars, Params).
parse_provider_args(['^'|Args], Derefs, I, [I|Vars], ['$VAR'(I)|Params]) :-
    I2 is I + 1,
    parse_provider_args(Args, Derefs, I2, Vars, Params).
parse_provider_args([X|_], _, _, _, _) :-
    error('invalid argument specification: ~w\n', [X]).

parse_provider(M:Call, Params, M:Caller) :-
    parse_provider(Call, Params, Caller).
parse_provider(foreign_call(Call), Params, foreign_call(Caller)) :-
    parse_provider(Call, Params, Caller).
parse_provider(Call, Params, Caller) :-
    Call =.. [Name|Args],
    append(Args, [Params], Args2),
    Caller =.. [Name|Args2].
parse_provider(Call, _, _) :-
    error('invalid provider call: ~w\n', [Call]).

find_include_file(Fname1, Fname) :-
    include_path(Dir),
    format_to_atom(Fname, '~a/~a', [Dir, Fname1]),
    file_exists(Fname), 
    !.
find_include_file(Fname, _) :-
    error('included file does not exist: ~a\n', [Fname]).

write_declarations(Out) :-
    declaration(Decl),
    format(Out, '-~q.\n', [Decl]),
    fail.
write_declarations(Out) :- 
    \+library_mode,
    format(Out, '-uses([lib, sys]).\n', []),
    fail.
write_declarations(Out) :- 
    findall(Mod, uses(Mod), Mods),
    format(Out, '-uses(~q).\n', [Mods]),
    fail.    
write_declarations(Out) :- nl(Out).

collect_groups([], []).
collect_groups([T|PRG], [[T|G]|GROUPS]) :-
    T = c(NA, _, _, _),
    filter_group(PRG, NA, NA, G, NG),
    collect_groups(NG, GROUPS).

filter_group([], _, _, [], []).
filter_group([T|PRG], NA, NA, [T|G], NG) :-
    T = c(NA, _, _, _),
    !,
    filter_group(PRG, NA, NA, G, NG).
filter_group([T|PRG], N/A, _, [T|G], NG) :-
    T = c(N/A, _, _, _),
    !,
    format(user_error, 'Warning: non-contiguous process definition for ~a/~d\n', [N, A]),
    filter_group(PRG, NA, NA, G, NG).
filter_group([T|PRG], NA, _, G, [T|NG]) :-
    T = c(NA2, _, _, _),    
    filter_group(PRG, NA, NA2, G, NG).

canonicalize((H :- G | B), N/A, Args, Guards, B) :-
    H =.. [N|Args],
    length(Args, A),
    serialize_guards(G, Guards).
canonicalize((H :- B), N/A, Args, Guards, B) :-
    canonicalize((H :- true | B), N/A, Args, Guards, B).
canonicalize(H, N/A, Args, Guards, B) :-
    canonicalize((H :- true | true), N/A, Args, Guards, B).

serialize_guards(true, []).
serialize_guards((true, G1), G) :- serialize_guards(G1, G).
serialize_guards((G1, G2), [G1|G3]) :- serialize_guards(G2, G3).
serialize_guards(G, [G]).

write_translation([], _).
write_translation([[]|Gs], Out) :-
    write_translation(Gs, Out).
write_translation([G|Gs], Out) :-
    write_group(G, Out),
    nl(Out),
    write_translation(Gs, Out).

write_group([], _).
write_group([c(N/_, Args, _, B)|G], Out) :-
    H =.. [N|Args],
    format(Out, '~q.\n', [(H :- B)]),
    write_group(G, Out).

% translation chain

translate(G, TG) :-
    check_definitions(G),
    check_overlap(G),
    inject_equality_guards(G, G2),
    expand_terms(G2, G3),
    eliminate_guards(G3, TG).

% check for library redefinitions

check_definitions(_) :- library_mode, !.
check_definitions(Gs) :-
    member(G, Gs), 
    check_definitions2(G), 
    fail.
check_definitions(_).

check_definitions2([c(NA, _, _, _)|_]) :-
    \+foreign(NA, _, _),
    \+library(NA, _),
    !.
check_definitions2([c(N/A, _, _, _)|_]) :-
    error('redefinition of library predicate ~a/~d\n', [N, A]).

% check argument overlap

check_overlap([]).
check_overlap([G|Gs]) :-
    check_group_overlap(G, 1, G),
    check_overlap(Gs).

check_group_overlap([], _, _).
check_group_overlap([c(_, [], _, _)|_], _, _) :- !.
check_group_overlap([c(_, Args, [], _)|_], I, G) :-
    find_overlap(G, Args, I, 1),
    !.
check_group_overlap([_|Cs], I, G) :-
    I2 is I + 1,
    check_group_overlap(Cs, I2, G).

find_overlap([], _, _, _) :- fail.
find_overlap([c(NA, Args1, [], _)|_], Args, I, J) :-
    I \= J,
    find_arg_overlap(Args1, Args, 1, 0, NA, I, J).
find_overlap([_|Cs], Args, I, J) :-
    J2 is J + 1,
    find_overlap(Cs, Args, I, J2).

find_arg_overlap([], _, _, 0, _, _, _) :- fail.
find_arg_overlap([], _, _, P, N/A, I, J) :- 
    P \= 0,
    format(user_error, 'Warning: clauses ~d and ~d of ~a/~d have overlapping match for arguments at position ~d but lack \'otherwise\' guard\n',
        [I, J, N, A, P]).
find_arg_overlap([X|_], [Y|_], _, _, _, _, _) :-
    discrete_value(X), 
    discrete_value(Y), 
    !, fail.
find_arg_overlap([X|_], [Y|_], _, _, _, _, _) :-
    \+congruent(X, Y),
    !, fail.
find_arg_overlap([X|Args1], [Y|Args2], P, 0, NA, I, J) :-
    structurally_same(X, Y),
    !,
    find_arg_overlap(Args1, Args2, P, P, NA, I, J).
find_arg_overlap([_|Args1], [_|Args2], P, O, NA, I, J) :-
    P2 is P + 1,
    find_arg_overlap(Args1, Args2, P2, O, NA, I, J).

discrete_value(X) :- atomic(X), !.
discrete_value('!'(_)) :- !.
discrete_value('$VAR'(_)) :- !, fail.
discrete_value([_|_]) :- !, fail.
discrete_value(X) :- X =.. [_, _|_].

congruent(X, _) :- atomic(X), !, fail.
congruent(_, Y) :- atomic(Y), !, fail.
congruent('$VAR'(_), _) :- !.
congruent(_, '$VAR'(_)) :- !.
congruent('!'(_), _) :- !.
congruent(_, '!'(_)) :- !.
congruent([H|T], Y) :-
    !,
    Y = [H2|T2],
    congruent(H, H2),
    congruent(T, T2).
congruent(X, Y) :- 
    X =.. [H|X2],
    Y =.. [H|Y2],
    congruent(X2, Y2).

structurally_same(X, X) :- !.
structurally_same('$VAR'(_), _) :- !.
structurally_same(_, '$VAR'(_)) :- !.
structurally_same(X, _) :- atomic(X), !, fail.
structurally_same(_, Y) :- atomic(Y), !, fail.
structurally_same([X1|X2], [Y1|Y2]) :-
    !,
    structurally_same(X1, Y1),
    structurally_same(X2, Y2).
structurally_same(X, Y) :-
    X =.. [HX|TX],
    Y =.. [HY|TY],
    structurally_same(HX, HY),
    structurally_same(TX, TY).
structurally_same(_, _).

% inject ==/2 guards for variables occurring multiple times in head

inject_equality_guards([], []).
inject_equality_guards([G1|Gs1], [G|Gs]) :-
    inject_equality_guards2(G1, G),
    inject_equality_guards(Gs1, Gs).

inject_equality_guards2([], []).
inject_equality_guards2([c(NA, Args1, Guards1, Body)|G1], [T|G]) :-
    T = c(NA, Args, Guards, Body),
    extract_variables((Args1, Body), Vars),
    duplicated_variables(Args1, Args, [], _, Vars, _, Guards1, Guards),
    inject_equality_guards2(G1, G).

duplicated_variables([], [], S, S, Vs, Vs, G, G).
duplicated_variables([Arg|Args], [Arg2|Args2], S1, S, Vs1, Vs, G1, G) :-
    duplicated_variables2(Arg, Arg2, S1, S2, Vs1, Vs2, G1, G2),
    duplicated_variables(Args, Args2, S2, S, Vs2, Vs, G2, G).

duplicated_variables2(X, X, S, S, Vs, Vs, G, G) :- atomic(X).
duplicated_variables2('$VAR'(V), '$VAR'(V2), S, S, Vs, [V2|Vs], G, [G1|G]) :-
    member(V, S),
    available_variable(Vs, V2),
    G1 = ('$VAR'(V) == '$VAR'(V2)).
duplicated_variables2('$VAR'(V), '$VAR'(V), S, [V|S], Vs, Vs, G, G).
duplicated_variables2(X, '$VAR'(NV), S, S, Vs, [NV|Vs], G, [G1|G]) :-
    ground_value(X),
    available_variable(Vs, NV),
    G1 = ('$VAR'(NV) == X).
duplicated_variables2([X1|Y1], [X|Y], S1, S, Vs1, Vs, G1, G) :-
    duplicated_variables2(X1, X, S1, S2, Vs1, Vs2, G1, G2),
    duplicated_variables2(Y1, Y, S2, S, Vs2, Vs, G2, G).
duplicated_variables2(T1, T, S1, S, Vs1, Vs, G1, G) :-
    callable(T1),
    T1 =.. [Name|Args1],
    duplicated_variables(Args1, Args, S1, S, Vs1, Vs, G1, G),
    T =.. [Name|Args].
duplicated_variables2(X, X, S, S, Vs, Vs, G, G).

% eliminate guards by introducing new groups

eliminate_guards([], []).
eliminate_guards([G1|Gs1], [G|Gs]) :-
    eliminate_guards1(G1, G),
    eliminate_guards(Gs1, Gs).

eliminate_guards1([c(NA, Args, Guards1, B)|G1], G) :-
    % 'otherwise' in first clause has no effect
    delete(Guards1, otherwise, Guards),
    eliminate_guards2([c(NA, Args, Guards, B)|G1], G).

eliminate_guards2([], []).
eliminate_guards2([C|Cs1], [C|Cs]) :-
    C = c(_, _, [], _),
    % no guards, try next clause in group
    eliminate_guards2(Cs1, Cs).
eliminate_guards2([C1|Cs1], Cs) :-
    C1 = c(NA, Args, Guards, B),
    member(otherwise, Guards),
    delete(Guards, otherwise, Guards2),
    eliminate_guards2([c(NA, Args, Guards2, B)|Cs1], [c(_, Args2, _, B2)|Cs3]),
    Cs = [c(NA, Args2, [], ('$suspend', B2))|Cs3].
eliminate_guards2([C1|Cs1], Cs) :-
    C1 = c(NA, Args, [data('$VAR'(V))|Guards], B),
    % data/1 guard
    force_variables(Args, [V], Args2),
    eliminate_guards2([c(NA, Args2, Guards, B)|Cs1], Cs).
eliminate_guards2([C1|Cs1], [C|Cs]) :-
    % clause with guards, excluding 'otherwise'
    C1 = c(NA, Args, Guards, B),
    C = c(NA, Args2, [], B2),
    extract_variables((Args, Guards, B), Vars),
    eliminate_guards3(Guards, Vars, Args, Args2, B, B2),
    eliminate_guards2(Cs1, Cs).

eliminate_guards3([], _, Args, Args, B, B).
eliminate_guards3([Guard|Guards], AllVars, Args, Args2, B1, B) :-
    extract_variables(Guard, HeadVars),
    available_variable(AllVars, HV),
    ResultVar = '$VAR'(HV),
    translate_guard(Guard, ResultVar, XGuard, [HV|AllVars], _, Force, Result),
    B = (XGuard, '$guard'(ResultVar, Result), B3),
    force_variables(Force, Args, HeadVars, Args3),
    eliminate_guards3(Guards, [HV|AllVars], Args3, Args2, B1, B3).

% utilities

rename_functor(N/A, P, Grist, RN) :-
    format_to_atom(RN, '$~d$~a$~a$~d', [P, Grist, N, A]).

extract_variables(X, Vs) :- extract_variables(X, Vs, []).

extract_variables(X, Vs, Vs) :- atomic(X).
extract_variables('$VAR'(V), [V|T], T).
extract_variables([X|Y], Vs, T) :-
    extract_variables(X, Vs, Vs2),
    extract_variables(Y, Vs2, T).
extract_variables(X, Vs, T) :-
    callable(X),
    X =.. [_|Args],
    extract_variables(Args, Vs, T).

force_variables(yes, X1, Vars, X) :- force_variables(X1, Vars, X).
force_variables(no, X, _, X).

force_variables(X, _, X) :- atomic(X).
force_variables('!'(X), _, '!'(X)).
force_variables('$VAR'(V), Vars, '!'('$VAR'(V))) :- member(V, Vars), !.
force_variables('$VAR'(V), _, '$VAR'(V)).
force_variables([X|Y], Vars, [X2|Y2]) :- 
    force_variables(X, Vars, X2),
    force_variables(Y, Vars, Y2).
force_variables(X, Vars, X2) :-
    callable(X),
    X =.. [N|Args],
    force_variables(Args, Vars, Args2),
    X2 =.. [N|Args2].

available_variable(Vars, V) :- available_variable(0, Vars, V).

available_variable(V, Vars, V) :- \+member(V, Vars), !.
available_variable(V1, Vars, V) :-
    V2 is V1 + 1,
    available_variable(V2, Vars, V).

ground_value(X) :- atomic(X), !.
ground_value([X|Y]) :- !, ground_value(X), ground_value(Y).
ground_value('$VAR'(_)) :- !, fail.
ground_value(T) :- 
    T =.. [_|Lst],
    ground_value(Lst).

% Guard translation

translate_guard(Exp, R, (X2, Y2, T), HV1, HV, Force, Result) :-
    callable(Exp),
    functor(Exp, Op, Arity),
    numeric_guard(Op/Arity),
    Exp =.. [_, X, Y],
    translate_expression(X, HV1, X2, XR, HV2),
    translate_expression(Y, HV2, Y2, YR, HV),
    Exp2 =.. [Op, XR, YR],
    translate_guard2(Exp2, R, T, Force, Result).
translate_guard(Exp, R, T, HV, HV, Force, Result) :-
    translate_guard2(Exp, R, T, Force, Result).

translate_guard2((X =:= Y), R, compute(=, X, Y, R), no, true).
translate_guard2('\\=:='(X, Y), R, compute(=, X, Y, R), no, false).
translate_guard2((X == Y), R, '$match'(X, Y, R), yes, true).
translate_guard2((X =\= Y), R, '$match'(X, Y, R), yes, false).
translate_guard2((X > Y), R, compute(>, X, Y, R), yes, true).
translate_guard2((X < Y), R, compute(<, X, Y, R), yes, true).
translate_guard2((X >= Y), R, compute(<, X, Y, R), yes, false).
translate_guard2((X =< Y), R, compute(>, X, Y, R), yes, false).
translate_guard2((X @< Y), R, compute(@<, X, Y, R), yes, true).
translate_guard2((X @> Y), R, compute(@>, X, Y, R), yes, true).
translate_guard2((X @>= Y), R, compute(@<, X, Y, R), yes, false).
translate_guard2((X @=< Y), R, compute(@>, X, Y, R), yes, false).
translate_guard2(idle, R, '$idle'(R), no, true).
translate_guard2(integer(X), R, '$integer'(X, R), yes, true).
translate_guard2(float(X), R, '$float'(X, R), yes, true).  % DEPRECATED
translate_guard2(real(X), R, '$float'(X, R), yes, true).
translate_guard2(number(X), R, '$number'(X, R), yes, true).
translate_guard2(string(X), R, '$atom'(X, R), yes, true).
translate_guard2(atomic(X), R, '$atomic'(X, R), yes, true).
translate_guard2(list(X), R, '$list'(X, R), yes, true).
translate_guard2(port(X), R, '$port'(X, R), yes, true).
translate_guard2(tuple(X), R, '$tuple'(X, R), yes, true).
translate_guard2(unknown(X), R, '$var'(X, R), no, true).
translate_guard2(known(X), R, '$var'(X, R), no, false).
translate_guard2(module(X), R, '$module'(X, R), yes, true).
translate_guard2(G, _, _, _, _) :- error('invalid guard: ~q\n', [G]).

opposite(false, true).
opposite(true, false).

translate_expression(rnd, HV, foreign_call(fl_random('$VAR'(V))), '$VAR'(V), [V|HV]) :- 
    !,
    available_variable(HV, V).
translate_expression(Exp, HV, true, Exp, HV) :- number(Exp).
translate_expression('$VAR'(V), HV, true, '$VAR'(V), HV).
translate_expression(-X, HV1, X2, XR, HV) :-
    translate_expression(0-X, HV1, X2, XR, HV).
translate_expression(+X, HV1, X2, XR, HV) :-
    translate_expression(X, HV1, X2, XR, HV).
translate_expression(\X, HV1, X2, XR, HV) :-
    translate_expression('><'(-1, X), HV1, X2, XR, HV).
translate_expression(Exp, HV1, (TX, TY, compute(NOp, RX, RY, '$VAR'(V))), '$VAR'(V), [V|HV3]) :-
    Exp =.. [Op, X, Y],
    numeric_operation(Op, NOp),
    translate_expression(X, HV1, TX, RX, HV2),
    translate_expression(Y, HV2, TY, RY, HV3),
    available_variable(HV3, V).
translate_expression(Exp, HV1, (TX, compute(NOp, RX, '$VAR'(V))), '$VAR'(V), [V|HV2]) :-
    Exp =.. [Op, X],
    numeric_operation(Op, NOp),
    translate_expression(X, HV1, TX, RX, HV2),
    available_variable(HV2, V).
translate_expression(Exp, _, _, _, _) :-
    error('invalid expression: ~w\n', [Exp]).

% expand is/2 expressions, explicit module calls and builtins:

expand_terms([], []).
expand_terms([G1|Gs1], [G|Gs]) :-
    expand_terms2(G1, G, Gs1, Gs2),
    expand_terms(Gs2, Gs).

expand_terms2([], [], Gs, Gs).
expand_terms2([C1|G1], [c(NA, Args, Guards, Body)|G], Gs1, Gs) :-
    C1 = c(NA, Args, Guards, Body1),
    extract_variables((Args, Guards, Body1), Vars),
    expand_body(Body1, Body, NA, Vars, _, Gs1, Gs2),
    expand_terms2(G1, G, Gs2, Gs).

counter(N) :-
    g_read(counter, N1),
    N is N1 + 1, 
    g_assign(counter, N).

%expand_body(X, _, _, _, _, _, _) :- dbg(X), fail.
expand_body((true, X), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars, Gs1, Gs).
expand_body((X, true), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars, Gs1, Gs).
expand_body('@'(Goal, Y), X2, NA, Vars1, Vars, Gs1, Gs) :-
    canonicalize_call_target(Goal, Goal2, Gs1, Gs2),
    expand_body('$rcall'(Goal2, Y), X2, NA, Vars1, Vars, Gs2, Gs).
expand_body('@'(X, Y), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body('$rcall'(X, Y), X2, NA, Vars1, Vars, Gs1, Gs).
expand_body(X = Y, '$unify_checked'(X, Y), _, Vs, Vs, Gs, Gs).
expand_body(':='(X, Y), '$assign'(Y, X), _, Vs, Vs, Gs, Gs).
expand_body(X is Exp, T, NA, Vars, [RV|Vars], Gs1, [G|Gs1]) :-
    counter(C),
    rename_functor(NA, C, x, N2),
    extract_variables(Exp, ExpVars),
    sort(ExpVars, SortedVars),
    length(SortedVars, SL),
    available_variable(Vars, RV),
    translate_expression(Exp, [RV|SortedVars], XBody, Result, _),
    build_exp_args(SortedVars, Args, Params, ['$VAR'(RV)]),
    Call =.. [N2|Params],
    T = (Call, (X = '$VAR'(RV))),
    SL2 is SL + 1,
    G = [c(N2/SL2, Args, [], (XBody, '$assign'(Result, '$VAR'(RV))))].
expand_body('&'(call(X, S), Y), T, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body((call(X, S), when(S, Y)), T, NA, Vars1, Vars, Gs1, Gs).
expand_body('&'(call(X, E, S),  Y), T, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body((call(X, E, S), when(S, Y)), T, NA, Vars1, Vars, Gs1, Gs).
expand_body('&'((X1, X2), Y), T, NA, Vars1, Vars, Gs1, Gs) :-
    wrap_term((X1, X2), NA, T2, Gs1, Gs2),
    expand_body('&'(T2, Y), T, NA, Vars1, Vars, Gs2, Gs).
expand_body('&'(X, Y), T, NA, Vars1, Vars, Gs1, Gs) :-
    callable(X),
    X =.. [N|Args],
    length(Args, A),
    confirmed_primitive(N/A),
    available_variable(Vars1, V),
    append(Args, ['$VAR'(V)], Args2),
    X2 =.. [N|Args2],
    expand_body((X2, when('$VAR'(V), Y)), T, NA, [V|Vars1], Vars, Gs1, Gs).
expand_body('&'(X, Y), T, NA, Vars1, Vars, Gs1, Gs) :-
    available_variable(Vars1, RV),
    expand_body((call(X, '$VAR'(RV)), when('$VAR'(RV), Y)), T, NA, [RV|Vars1], Vars, 
        Gs1, Gs).
expand_body(when(X, T), T2, NA, Vars, Vars, Gs1, [G|Gs1]) :-
    check_when(X, T, V),
    counter(C),
    rename_functor(NA, C, w, N2),
    extract_variables((X, T), WVars),
    sort(WVars, SortedVars),
    length(SortedVars, SL),
    build_when_args(SortedVars, V, Args, Params),
    T2 =.. [N2|Params],
    G = [c(N2/SL, Args, [], T)].
expand_body(fmt:format(File, Fstr, Args, Done), X, NA, Vars1, Vars, Gs1, Gs) :-
    add_use(fmt),
    compile_format(Fstr, Args, NA, Items),
    expand_body(fmt:format_chunked(File, Items, Done), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(fmt:format(Fstr, Args, Done), X, NA, Vars1, Vars, Gs1, Gs) :-
    add_use(fmt),
    compile_format(Fstr, Args, NA, Items),
    expand_body(fmt:format_chunked(1, Items, Done), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(fmt:format(Fstr, Args), X, NA, Vars1, Vars, Gs1, Gs) :-
    add_use(fmt),
    compile_format(Fstr, Args, NA, Items),
    available_variable(Vars1, RV),
    expand_body(fmt:format_chunked(1, Items, '$VAR'(RV)), X, NA, [RV|Vars1], 
        Vars, Gs1, Gs).
expand_body(fmt:format_chars(Fstr, Args, Out, Tail), X, NA, Vars1, Vars, Gs1, Gs) :-
    add_use(fmt),
    compile_format(Fstr, Args, NA, Items),
    expand_body(fmt:outputs(Items, Out, Tail), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(fmt:format_chars(Fstr, Args, Out), X, NA, Vars1, Vars, Gs1, Gs) :-
    add_use(fmt),
    compile_format(Fstr, Args, NA, Items),
    expand_body(fmt:outputs(Items, Out, []), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(M:T, X, NA, Vars1, Vars, Gs1, Gs) :-
    functor(T, N, A),
    applicative(M:N/A),
    T =.. [_, Goal|Args],
    canonicalize_call_target(Goal, Goal2, Gs1, Gs2),
    T2 =.. [N, Goal2|Args],
    expand_body(M:T2, X, NA, Vars1, Vars, Gs2, Gs).    
expand_body(M:T, X, NA, Vars1, Vars, Gs1, Gs) :-
    check_module(M, M:T),
    add_use(M),
    expand_body(call(M:T), X, NA, Vars1, Vars, Gs1, Gs).
expand_body('$VAR'(V), X, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(call('$VAR'(V)), X, NA, Vars1, Vars, Gs1, Gs).
expand_body(call(T, S), T2, NA, Vars, Vars, Gs, [G|Gs]) :-
    maybe_add_use(T),
    local_call(T),
    counter(C),
    rename_functor(NA, C, l, N2),
    extract_variables((T, S), LVars),
    sort(LVars, SortedVars),
    length(SortedVars, SL),
    var_lists(SortedVars, [], Args2, Params),
    T2 =.. [N2|Params],
    available_variable(SortedVars, V1),
    G = [c(N2/SL, Args2, [], (foreign_call(fl_new_task(S, '$VAR'(V1))), T))].
expand_body(call(T, E, S), T2, NA, Vars, Vars, Gs, [G|Gs]) :-
    maybe_add_use(T),
    local_call(T),
    counter(C),
    rename_functor(NA, C, l, N2),
    extract_variables((T, S, E), LVars),
    sort(LVars, SortedVars),
    length(SortedVars, SL),
    var_lists(SortedVars, [], Args2, Params),
    T2 =.. [N2|Params],
    available_variable(SortedVars, V1),
    G = [c(N2/SL, Args2, [], (foreign_call(fl_new_task(E, S, '$VAR'(V1))), T))].
expand_body(T, X, NA, Vars1, Vars, Gs1, Gs) :-
    functor(T, N, A),
    applicative(N/A),
    T =.. [_, Goal|Args],
    canonicalize_call_target(Goal, Goal2, Gs1, Gs2),
    T2 =.. [N, Goal2|Args],
    expand_body(T2, X, NA, Vars1, Vars, Gs2, Gs).    
expand_body(apply(T, Args), X, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(lib:'$apply'(T, Args, '$VAR'('$MODULE')), X, NA, Vars1, Vars, Gs1, Gs).
expand_body((X, Y), (X2, Y2), NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars2, Gs1, Gs2),
    expand_body(Y, Y2, NA, Vars2, Vars, Gs2, Gs).
expand_body(T, T2, NA, Vars, Vars, Gs1, Gs) :-
    callable(T),
    functor(T, NB, A),
    T =.. [_|Args],
    foreign(NB/A, Op, Spec),
    expand_builtin_call(Spec, Args, Op, NA, T2, Gs1, Gs).
expand_body(T, T2, _, Vars, Vars, Gs, Gs) :-
    callable(T),
    functor(T, NB, A),
    \+library_mode,
    T =.. [_|Args],
    library(NB/A, M:Call),
    C =.. [Call|Args],
    T2 = call(M:C).
expand_body(T, T, _, Vs, Vs, Gs, Gs).

% convert goal to callable user goal or fully qualified library call, or fail:
canonicalize_call_target('@'(X, Y), Goal, Gs1, Gs) :-
    wrap_primitive('$rcall'(X, Y), Goal, Gs1, Gs).
canonicalize_call_target(Goal1, Goal, Gs1, Gs) :-
    primitive_call(Goal1),
    wrap_primitive(Goal1, Goal, Gs1, Gs).
canonicalize_call_target(Goal1, GM:Goal, Gs, Gs) :-
    primitive_library_call(Goal1, GM:GF),
    Goal1 =.. [_|GArgs],
    Goal =.. [GF|GArgs].

wrap_primitive(T, T2, Gs, [G|Gs]) :-
    counter(C),
    T =.. [N|Args],
    length(Args, A),
    extract_variables(T, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    rename_functor(N/A, C, p, N2),
    var_lists(SortedVars, [], Args2, Params),
    T2 =.. [N2|Args2],
    Call =.. [N|Args],
    G = [c(N2/SL, Params, [], Call)].

wrap_term(T, NA, T2, Gs, [G|Gs]) :-
    counter(C),
    extract_variables(T, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    rename_functor(NA, C, t, N2),
    var_lists(SortedVars, [], Args2, _),
    T2 =.. [N2|Args2],
    G = [c(N2/SL, Args2, [], T)].

local_call(T) :-
    T \= _:_,
    \+primitive_call(T), 
    \+primitive_library_call(T, _).

check_when('$VAR'(V), _, V).
check_when(X, T, _) :- error('invalid argument to `when\': ~q\n', [when(X, T)]).

check_module(M, _) :- atom(M), !.
check_module(_, X) :- error('invalid module prefix in `~q\'\n', [X]).

build_when_args([], _, [], []).
build_when_args([V|Vs], V, ['!'('$VAR'(V))|Args], ['$VAR'(V)|Params]) :-
    build_when_args(Vs, V, Args, Params).
build_when_args([V1|Vs], V, ['$VAR'(V1)|Args], ['$VAR'(V1)|Params]) :-
    build_when_args(Vs, V, Args, Params).

expand_builtin_call(simple, Args, Op, _, R, Gs, Gs) :-
    T =.. [Op|Args],
    R = foreign_call(T).
expand_builtin_call(Spec, Args, Op, NA, R, Gs, [G|Gs]) :-
    counter(C),
    rename_functor(NA, C, f, N2),
    extract_variables(Args, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    scan_builtin_args(Spec, Args, Forced),
    var_lists(SortedVars, Forced, Args2, Params),
    T =.. [Op|Args],
    R =.. [N2|Params],
    G = [c(N2/SL, Args2, [], foreign_call(T))].

scan_builtin_args([], _, []).
scan_builtin_args([f|Spec], ['$VAR'(V)|Args], [V|Forced]) :-
    scan_builtin_args(Spec, Args, Forced).
scan_builtin_args([_|Spec], [_|Args], Forced) :-
    scan_builtin_args(Spec, Args, Forced).

var_lists([], _, [], []).
var_lists([V|Vs], Forced, ['!'('$VAR'(V))|Args], ['$VAR'(V)|Params]) :-
    member(V, Forced),
    var_lists(Vs, Forced, Args, Params).
var_lists([V|Vs], Forced, ['$VAR'(V)|Args], ['$VAR'(V)|Params]) :-
    var_lists(Vs, Forced, Args, Params).

build_exp_args([], T, T, T).
build_exp_args([V|Vs], ['!'('$VAR'(V))|XVs], ['$VAR'(V)|PVs], T) :-
    build_exp_args(Vs, XVs, PVs, T).

% precompilation of format strings

compile_format([], [], _, []).
compile_format([], Args, N/A, []) :-
    warning('~a/~d: extra arguments in format string: ~w\n', [N, A, Args]).
compile_format(Fstr, Args, NA, Out) :-
    atom(Fstr), 
    atom_codes(Fstr, Str),
    compile_format(Str, Args, NA, Out).
compile_format(Fstr, _, _, _) :-
    \+ground_format_string(Fstr),
    !, fail.
compile_format(_, Args, _, _) :-
    \+valid_format_args(Args),
    !, fail.
compile_format([0'~, 0'~|Fstr], Args, NA, [0'~|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format([0'~, 0'n|Fstr], Args, NA, [10|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format([0'~, C|Fstr], Args,NA,  Out) :-
    compile_format_char(C, Fstr, Args, NA, Out).
compile_format(Fstr, Args, NA, [S|Out]) :-
    compile_format_chunk(Fstr, Fstr2, CL),
    atom_codes(S, CL),
    compile_format(Fstr2, Args, NA, Out).

compile_format_char(C, Fstr, [], N/A, [0'~, C|Out]) :-
    warning('~a/~d: missing argument for ~~~c in format string: ...~s\n', 
        [N, A, C, Fstr]),
    compile_format(Fstr, [], N/A, Out).
compile_format_char(0'c, Fstr, [X|Args], NA, [X|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format_char(0'?, Fstr, [F, Args1|Args], NA, [f(F, Args1)|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format_char(0'?, Fstr, _, N/A, [0'~, 0'?|Out]) :-
    warning('~a/~d: missing argument for ~~? in format string: ...~s\n', 
        [N, A, Fstr]),
    compile_format(Fstr, [], N/A, Out).
compile_format_char(C, Fstr, [X|Args], NA, [Item|Out]) :-
    member(C, "dxasqw"),
    atom_codes(N, [C]),
    Item =.. [N, X],
    compile_format(Fstr, Args, NA, Out).
compile_format_char(C, Fstr, Args, N/A, [0'~, C|Out]) :-
    warning('~a/~d: invalid formatting directive in format string: ...~s\n', 
        [N, A, [C|Fstr]]),
    compile_format(Fstr, Args, N/A, Out).

compile_format_chunk([], [], []).
compile_format_chunk([0'~|Fstr], [0'~|Fstr], []).    
compile_format_chunk([C|Fstr1], Fstr, [C|S]) :-
    compile_format_chunk(Fstr1, Fstr, S).

ground_format_string([]).
ground_format_string([C|X]) :-
    integer(C), 
    ground_format_string(X).

valid_format_args([]).
valid_format_args([_|Args]) :- valid_format_args(Args).

% tables

primitive_call(_=_).
primitive_call('$rcall'(_, _)).
primitive_call(call(_)).
primitive_call(run(_, _)).
primitive_call(call(_, _)).
primitive_call(call(_, _, _)).
primitive_call(unify(_, _, _)).
primitive_call(compute(_, _, _)).
primitive_call(compute(_, _, _, _)).
primitive_call(foreign_call(_)).
primitive_call(Term) :-
    functor(Term, Name, Arity),
    foreign(Name/Arity, _, _).

primitive_library_call(Term, Libcall) :-
    functor(Term, Name, Arity),
    library(Name/Arity, Libcall).

confirmed_primitive(assign/2).
confirmed_primitive(chdir/1).
confirmed_primitive(chmod/2).
confirmed_primitive(delete_file/1).
confirmed_primitive(deref/1).
confirmed_primitive(get_arg/3).
confirmed_primitive(get_global/2).
confirmed_primitive(log/1).
confirmed_primitive(put_arg/2).
confirmed_primitive(put_global/2).
confirmed_primitive(randomize/1).
confirmed_primitive(rmdir/1).
confirmed_primitive(send/2).
confirmed_primitive(writeln/1).

foreign(command_line/1, fl_command_line, simple).
foreign(open_port/2, fl_open_port, simple).
foreign(assign/2, fl_assign, simple).
foreign(assign/3, fl_assign, simple).
foreign(threads/1, fl_threads, simple).
foreign(getpid/1, fl_getpid, simple).
foreign(self/1, fl_self, simple).
foreign(current_seconds/1, fl_current_seconds, simple).
foreign(getcwd/1, fl_getcwd, simple).
foreign(idle_thread/1, fl_idle_peer, simple).
foreign(environment/1, fl_environment, simple).
foreign(all_modules/1, fl_all_modules, simple).
foreign(string_to_integer/2, fl_string_to_integer, [f, x]).
foreign(string_to_integer/3, fl_string_to_integer, [f, f, x]).
foreign(string_to_real/2, fl_string_to_float, [f, x]).
foreign(string_to_float/2, fl_string_to_float, [f, x]). % DEPRECATED
foreign(string_to_list/3, fl_string_to_list, [f, x, x]).
foreign(string_to_byte_list/3, fl_string_to_byte_list, [f, x, x]).
foreign(number_to_list/3, fl_number_to_list, [f, x, x]).
foreign(number_to_list/4, fl_number_to_list, [f, f, x, x]).
foreign(close_file/1, fl_close_file, [f]).
foreign(close_file/2, fl_close_file, [f, x]).
foreign(utf_encode/3, fl_utf_encode, [f, x, x]).
foreign(send/2, fl_send, [f, x]).
foreign(send/3, fl_send, [f, x, x]).
foreign(tuple_to_list/3, fl_tuple_to_list, [f, x, x]).
foreign(put_arg/3, fl_put_arg, [f, f, x]).
foreign(put_arg/4, fl_put_arg, [f, f, x, x]).
foreign(get_arg/3, fl_get_arg, [f, f, x]).
foreign(get_arg/4, fl_get_arg, [f, f, x, x]).
foreign(make_tuple/2, fl_make_tuple, [f, x]).
foreign(halt/1, fl_halt, [f]).
foreign(listen/2, fl_listen, [f, x]).
foreign(signal/2, fl_signal, [f, x]).
foreign(signal/3, fl_signal, [f, x, x]).
foreign(kill/2, fl_kill, [f, f]).
foreign(kill/3, fl_kill, [f, f, x]).
foreign(timer/3, fl_timer, [f, x, x]).
foreign(clock/3, fl_clock, [f, x, x]).
foreign(cancel_timer/1, fl_cancel_timer, [f]).
foreign(cancel_timer/2, fl_cancel_timer, [f, x]).
foreign(restart_timer/2, fl_restart_timer, [f, f]).
foreign(restart_timer/3, fl_restart_timer, [f, f, x]).
foreign(get_global/3, fl_get_global, [f, x, x]).
foreign(get_global/2, fl_get_global, [f, x]).
foreign(put_global/2, fl_put_global, [f, x]).
foreign(put_global/3, fl_put_global, [f, x, x]).
foreign(compare/3, fl_compare, [f, f, x]).
foreign(get_module/2, fl_get_module, [f, x]).
foreign(module_name/2, fl_module_name, [f, x]).
foreign(module_exports/2, fl_module_exports, [f, x]).

library(length/2, lib:length).
library(deref/1, lib:deref).
library(deref/2, lib:deref).
library(error/1, lib:error).
library(writeln/1, lib:writeln).
library(writeln/2, lib:writeln).
library(list_to_string/2, lib:list_to_string).
library(list_to_tuple/2, lib:list_to_tuple).
library(list_to_number/2, lib:list_to_number).
library(list_to_number/3, lib:list_to_number).
library(write_file/3, lib:write_file).
library(write_file/4, lib:write_file).
library(read_file/4, lib:read_file).
library(utf_encode/2, lib:utf_encode).
library(utf_encode/3, lib:utf_encode).
library(utf_decode/2, lib:utf_decode).
library(utf_decode/3, lib:utf_decode).
library(append/3, lib:append).
library(append/2, lib:append).
library(member/3, lib:member).
library(log/1, lib:log).
library(log/2, lib:log).
library(chdir/1, lib:chdir).
library(chdir/2, lib:chdir).
library(getenv/2, lib:getenv).
library(file_type/2, lib:file_type).
library(file_size/2, lib:file_size).
library(readlink/2, lib:readlink).
library(open_file/3, lib:open_file).
library(delete_file/1, lib:delete_file).
library(delete_file/2, lib:delete_file).
library(rmdir/1, lib:rmdir).
library(rmdir/2, lib:rmdir).
library(chmod/2, lib:chmod).
library(chmod/3, lib:chmod).
library(file_exists/2, lib:file_exists).
library(environment/1, lib:environment).
library(merger/2, lib:merger).
library(real_to_list/2, lib:real_to_list).
library(list_to_real/2, lib:list_to_real).
library(randomize/1, lib:randomize).
library(randomize/2, lib:randomize).

numeric_guard(('=:=')/2).
numeric_guard(('\\=:=')/2).
numeric_guard(('>')/2).
numeric_guard(('<')/2).
numeric_guard(('>=')/2).
numeric_guard(('=<')/2).

numeric_operation('/\\', and).
numeric_operation('\\/', or).
numeric_operation('><', xor).
numeric_operation('<<', shl).
numeric_operation('>>', shr).
numeric_operation('\\\\', mod).
numeric_operation(Op, Op) :- numeric_operation(Op).

numeric_operation(+).
numeric_operation(-).
numeric_operation(*).
numeric_operation(/).
numeric_operation(sqrt).
numeric_operation(integer).
numeric_operation(real).
numeric_operation(float). % DEPRECATED
numeric_operation(floor).
numeric_operation(ceiling).
numeric_operation(round).
numeric_operation(truncate).
numeric_operation(max).
numeric_operation(min).
numeric_operation(real_fractional_part).
numeric_operation(real_integer_part).
numeric_operation(log).
numeric_operation(exp).
numeric_operation(sign).
numeric_operation(sin).
numeric_operation(cos).
numeric_operation(tan).
numeric_operation(atan).
numeric_operation(**).
numeric_operation(rnd).
numeric_operation(abs).

applicative(call/1).
applicative(call/2).
applicative(call/3).
applicative('$rcall'/2).
applicative(app:maplist/3).
applicative(app:foreach/3).
applicative(app:filter/3).
applicative(app:partition/4).
applicative(app:foldl/4).
applicative(app:mapreduce/4).
