/* FLENG - Linux specific part of the runtime system

  Copyright (c) 2021, Felix L. Winkelmann
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. The name of the authors may not be used to endorse or promote products
     derived from this software without specific prior written permission.
  
  THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "fleng.h"
#include "fleng-util.h"
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/signalfd.h>
#include <sys/timerfd.h>
#include <sys/wait.h>
#include <string.h>
#include <pthread.h>
#include <poll.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>

#define MAX_EVENTS  32

static struct { char *name; int sig; } signals[] = {
  { "SIGHUP", 1 },
  { "SIGINT", 2 },
  { "SIGQUIT", 3 },
  { "SIGILL", 4 },
  { "SIGTRAP", 5 },
  { "SIGABRT", 6 },
  { "SIGIOT", 6 },
  { "SIGFPE", 8 },
  { "SIGKILL", 9 },
  { "SIGSEGV", 11 },
  { "SIGPIPE", 13 },
  { "SIGALRM", 14 },
  { "SIGTERM", 15 },
  { "SIGTTIN", 21 },
  { "SIGTTOU", 22 },
  { "SIGXCPU", 24 },
  { "SIGXFSZ", 25 },
  { "SIGVTALRM", 26 },
  { "SIGPROF", 27 },
  { "SIGWINCH", 28 },
  { "SIGBUS", 7 },
  { "SIGUSR1", 10 },
  { "SIGCHLD", 17 },
  { "SIGCONT", 18 },
  { "SIGIO", 29 },
  { "SIGSTOP", 19 },
  { "SIGSYS", 31 },
  { "SIGTSTP", 20 },
  { "SIGURG", 23 },
  { "SIGUSR2", 12 },
  { "SIGSTKSZ", 0x2000 },
  { "SIGPOLL", 29 },
  { "SIGPWR", 30 },
  { "SIGSTKFLT", 16 },
  { "SIGUNUSED", 31 },
  { "SIGRTMAX", 0 },
  { "SIGRTMIN", 0 },
  { NULL, 0 }
};

struct child {
    pid_t pid;
    FL_TCB *tcb;
    FL_VAL var;
    int running;
    int status;
    struct child *next;
};

static volatile int sigchild = 0, sigint = 0, sigterm = 0;
static volatile int sigint_enabled = 1, sigterm_enabled = 1;
static pthread_mutex_t child_mutex = PTHREAD_MUTEX_INITIALIZER;
static struct child *children;

int fl_resolve_signal(char *name) 
{
    for(int i = 0; signals[ i ].name != NULL; ++i) {
        if(!strcmp(signals[ i ].name, name)) 
            return(signals[ i ].sig);
    }
    return -1;
}

static void signal_handler(int sig) 
{
    switch(sig) {
    case SIGCHLD: 
        sigchild = 1;
        break;
    case SIGINT:
        sigint = 1;
        break;
    case SIGTERM:
        sigterm = 1;
        break;
    }
}

int fl_create_event_queue(FL_TCB *tcb)
{
    tcb->pollfds = (struct pollfd *)malloc(sizeof(struct pollfd) * MAX_EVENTS);
    struct sigaction sa;
    sa.sa_handler = signal_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGCHLD, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigset_t ss;
    sigfillset(&ss);
    sigdelset(&ss, SIGINT);
    sigdelset(&ss, SIGCHLD);
    sigdelset(&ss, SIGTERM);
    sigdelset(&ss, SIGPROF);
    pthread_sigmask(SIG_BLOCK, &ss, NULL);
    children = NULL;
    return 0;
}

static void wake_children(FL_TCB *tcb) 
{
    if(sigchild) {
        sigchild = 0;
        for(;;) {
            int s;
            pid_t w = waitpid(-1, &s, WNOHANG);
            if(w <= 0) break;
            pthread_mutex_lock(&child_mutex);
            for(struct child *c = children; c != NULL; c = c->next) {
                if(c->pid == w) {
                    c->running = 0;
                    if(!WIFEXITED(s)) s = 255; /* XXX ? */
                    else s = WEXITSTATUS(s);
                    c->status = s;
                }
            }
            pthread_mutex_unlock(&child_mutex);
        }
    }
    if(children == NULL) return;
    pthread_mutex_lock(&child_mutex);
    struct child **prev = &children;
    for(struct child *c = children; c != NULL; c = c->next) {
        if(c->tcb == tcb && !c->running) {
            fl_assign(tcb, MKINT(c->status), c->var);
            /* var is kept in listeners, no need to unref */
            *prev = c->next;
            free(c);
        } else prev = &c->next;
    }
    pthread_mutex_unlock(&child_mutex);
}

void fl_check_interrupts(FL_TCB *tcb) 
{
    if(sigint && sigint_enabled) fl_terminate(2);
    if(sigterm && sigterm_enabled) fl_terminate(2);
    wake_children(tcb);
}

static void assign_tail(FL_TCB *tcb, FL_VAL lrec, FL_VAL val) 
{
    FL_VAL var = CAR(CDR(lrec));
    FL_VAL t = addref(fl_chase_tail(var));
    FL_VAL newt = mkvar(tcb);
    fl_assign(tcb, mklist(tcb, val, newt), t);
    unref(tcb, var);
    CAR(CDR(lrec)) = t;
}

static void listeners(FL_TCB *tcb, char *msg)
{
    fprintf(stderr, "%s: ", msg);
    for(FL_VAL ll = tcb->listening; ll != fl_nil; ll = CDR(ll)) {
        FL_VAL lrec = CAR(ll);
        fprintf(stderr, "fd=%ld, t=%ld; ", INT(CAR(lrec)), INT(CAR(CDR(CDR(lrec)))));
    }
    fputs("\n", stderr);
}

static int update_events(FL_TCB *tcb, long ms)
{
    int nt = 0;
    struct pollfd *pollfds = tcb->pollfds;
    /* XXX re-use pollfds if listeners didn't change */
    for(FL_VAL ll = tcb->listening; ll != fl_nil; ll = CDR(ll)) {
        assert(nt < MAX_EVENTS);
        FL_VAL litem = CAR(ll);
        int fd = INT(CAR(litem));
        if(INT(CAR(CDR(CDR(litem)))) != FL_CHILD) {
            pollfds[ nt ].fd = fd;
            pollfds[ nt ].events = POLLIN;
            ++nt;
        }
    }
    int r = poll(pollfds, nt, ms);
    if(r == -1 && errno != EINTR) fl_rt_error(tcb, fl_nil, FL_IO_ERROR);
    int intr = errno == EINTR;
    fl_check_interrupts(tcb);
    FL_VAL ll = tcb->listening; 
    FL_VAL *prev = &tcb->listening;
    while(ll != fl_nil) {
        FL_VAL lrec = CAR(ll);
        FL_VAL next = CDR(ll);
        int fd = INT(CAR(lrec));
        int event = INT(CAR(CDR(CDR(lrec))));
        FL_VAL var = CAR(CDR(lrec));
        if(event == FL_CHILD) {
            if(CAR(var) != var) { /* bound by wake_children() */
                *prev = next;
                CDR(ll) = fl_nil;
                unref(tcb, ll);
            } else prev = &CDR(ll);
        } else if(!intr) {
            int found = 0;
            for(int i = 0; i < nt; ++i) {
                int drop = 1;
                if(fd == pollfds[ i ].fd && 
                   (pollfds[ i ].revents & (POLLIN | POLLHUP)) != 0) {
                    uint64_t n = 0;
                    struct signalfd_siginfo si;
                    found = 1;
                    switch(event) {
                    case FL_TIMEOUT:
                        read(fd, &n, sizeof(uint64_t));
                        if(CAR(var) == var) fl_assign(tcb, fl_nil, var);
                        close(fd);
                        break;
                    case FL_CLOCK:
                        read(fd, &n, sizeof(uint64_t));
                        assign_tail(tcb, lrec, MKINT(n));
                        drop = 0;
                        break;
                    case FL_SIGNAL:
                        read(fd, &si, sizeof(struct signalfd_siginfo));
                        assign_tail(tcb, lrec, MKINT(1));
                        drop = 0;
                        break;
                    case FL_INPUT:
                        fl_assign(tcb, fl_nil, var);
                        break;
                    }
                    if(drop) {
                        *prev = next;
                        CDR(ll) = fl_nil;
                        unref(tcb, ll);
                    } 
                }
            }
            if(!found) prev = &CDR(ll);
        } else prev = &CDR(ll);
        ll = next;
    }
    if(intr) return 0;
    return r;
}

int fl_wait_for_events(FL_TCB *tcb, long ms)
{
    return update_events(tcb, ms);
}

int fl_add_event_to_queue(FL_TCB *tcb, int event, FL_VAL data, FL_VAL var, FL_VAL eid)
{
    /* listener record: [fd, var, event|data] */
    FL_VAL n1 = mklist(tcb, MKINT(event), data);
    n1 = mklist(tcb, var, n1);
    n1 = mklist(tcb, fl_nil, n1);
    FL_VAL n = mklist(tcb, n1, tcb->listening);
    unref(tcb, tcb->listening);
    tcb->listening = addref(n);
    sigset_t sigs;
    int fd;
    switch(event) {
    case FL_INPUT:  /* data: fd */
        fd = INT(data);
        break;
    case FL_SIGNAL: { /* data: signal */
        int s = INT(data);
        sigemptyset(&sigs);
        sigaddset(&sigs, s);
        if(s == SIGINT) sigint_enabled = 0;
        if(s == SIGTERM) sigterm_enabled = 0;
        fd = signalfd(-1, &sigs, SFD_CLOEXEC);
        break; }
    case FL_TIMEOUT:    /* data: milliseconds */
    case FL_CLOCK: {
        struct itimerspec it;
        fd = timerfd_create(CLOCK_MONOTONIC, TFD_CLOEXEC);
        if(fd < 0) fl_rt_error(tcb, data, FL_IO_ERROR);
        int ms = INT(data);
        it.it_value.tv_sec = ms / 1000;
        it.it_value.tv_nsec = (ms % 1000) * 1000000;
        if(event == FL_CLOCK) it.it_interval = it.it_value; 
        else it.it_interval.tv_sec = it.it_interval.tv_nsec = 0;
        if(timerfd_settime(fd, 0, &it, NULL) < 0) 
            fl_rt_error(tcb, MKINT(fd), FL_IO_ERROR);
        break; }
    case FL_CHILD: { /* data: pid */
        struct child *c = (struct child *)malloc(sizeof(struct child));
        c->running = 1;
        c->tcb = tcb;
        c->pid = INT(data);
        c->var = var;
        pthread_mutex_lock(&child_mutex); /* XXX use atomic */
        c->next = children;
        children = c;
        pthread_mutex_unlock(&child_mutex);
        fd = 0;
        break; }
    }
    if(fd == - 1) {
        tcb->listening = CDR(n);
        CDR(n) = fl_nil;
        unref(tcb, n); 
        fl_rt_error(tcb, data, FL_IO_ERROR);
        return 0;
    }
    CAR(n1) = MKINT(fd);
    if(eid != NULL) fl_assign(tcb, MKINT(fd), eid);
    return update_events(tcb, 0);
}

FL_VAL find_listener(FL_TCB *tcb, FL_VAL id, int remove)
{
    FL_VAL *prev = &tcb->listening;
    FL_VAL ll = tcb->listening;
    while(ll != fl_nil) {
        FL_VAL lrec = CAR(ll);
        if(CAR(lrec) == id) {
            if(remove) {
                addref(lrec);
                *prev = CDR(ll);
                CDR(ll) = fl_nil;
                unref(tcb, ll);
            }
            return lrec;
        }
        prev = &CDR(ll);
        ll = CDR(ll);
    }
    return fl_nil;
}

int fl_cancel_timer(FL_TCB *tcb, FL_VAL id)
{
    FL_VAL *prev;
    FL_VAL lrec = find_listener(tcb, id, 1);
    if(lrec == fl_nil) return 0;
    FL_VAL var = CAR(CDR(lrec));
    if(INT(CAR(CDR(CDR(lrec)))) == FL_CLOCK) {
        FL_VAL st = fl_chase_tail(var);
        if(BITS(st) == 0 && TAG(st) == FL_VAR_TAG) 
            fl_assign(tcb, fl_nil, st);
    } else if(CAR(var) == var) fl_assign(tcb, fl_false, var);
    unref(tcb, lrec);
    close(INT(CAR(lrec)));
    return 1;
}

int fl_restart_timer(FL_TCB *tcb, FL_VAL id, int ms)
{
    FL_VAL *prev;
    FL_VAL lrec = find_listener(tcb, id, 0);
    if(lrec == fl_nil) return 0;
    struct itimerspec it;
    it.it_value.tv_sec = ms / 1000;
    it.it_value.tv_nsec = (ms % 1000) * 1000000;
    if(INT(CAR(CDR(CDR(lrec)))) == FL_CLOCK)
        it.it_interval = it.it_value; 
    else it.it_interval.tv_sec = it.it_interval.tv_nsec = 0;
    int fd = INT(CAR(lrec));
    if(timerfd_settime(fd, 0, &it, NULL) < 0) 
        fl_rt_error(tcb, MKINT(fd), FL_IO_ERROR);
    return 1;
}
